/*
   Move the definitions of the Comparator classes
   into the scope of the main() method as inner
   nested classes.
*/
import java.util.Comparator;
import java.util.Arrays;
import java.util.List;

public class Comparator_two_functor_v3
{
   public static void main(String[] args)
   {
      List<Country> countries = Arrays.asList(
         new Country("Uruguay",   176220,   3.4),
         new Country("Thailand",  514000,  67.0),
         new Country("Belgium",    30528,  11.0),
         new Country("Zaire",    2345410,  46.5),
         new Country("Belize",     22966,   0.334));

      // Since we are not using lambdas, and we cannot pass operators
      // as parameters in Java, we need the following interfaces as
      // wrappers for simple functions.
      abstract class Fn<T> { // create a "functor" type
         public abstract boolean fn(T x, T y);
      }

      class FnLT extends Fn<Double> {
         public boolean fn(Double x, Double y) {
            return x < y;
         }
      }

      class FnGT extends Fn<Double> {
         public boolean fn(Double x, Double y) {
            return x > y;
         }
      }

      class FnStringLT extends Fn<String> {
         @Override
         public boolean fn(String x, String y) {
            return x.compareTo(y) <= -1;
         }
      }

      class FnStringGT extends Fn<String> {
         @Override
         public boolean fn(String x, String y) {
            return x.compareTo(y) >= +1;
         }
      }


      abstract class Get<T> { // create a "functor" type
         public abstract T get(Country c);
      }

      class GetArea extends Get<Double> {
         @Override
         public Double get(Country c) {
            return c.getArea();
         }
      }

      class GetPopulation extends Get<Double> {
         @Override
         public Double get(Country c) {
            return c.getPopulation();
         }
      }

      class GetName extends Get<String> {
         @Override
         public String get(Country c) {
            return c.getName();
         }
      }

      // This comparator is now a class nested inside of main().
      class CountryComparator<T> implements Comparator<Country> {
         Fn<T> fn;    // functor that holds <, >, <=, or >=
         Get<T> get;  // get method

         CountryComparator(Fn<T> fn, Get<T> get) {
            this.fn = fn;
            this.get = get;
         }

         public int compare(Country c1, Country c2) {
            if ( fn.fn(get.get(c1), get.get(c2)) ) return -1;
            if (!fn.fn(get.get(c1), get.get(c2)) ) return  1;
            return 0;
         }
      }


      // Build the comparison functors using the two parameter comparison factory.
      countries.sort(new CountryComparator<Double>(new FnLT(), new GetArea()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparator<Double>(new FnGT(), new GetArea()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparator<Double>(new FnLT(), new GetPopulation()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparator<Double>(new FnGT(), new GetPopulation()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparator<String>(new FnStringLT(), new GetName()));
      countries.forEach(System.out::println);System.out.println();

      countries.sort(new CountryComparator<String>(new FnStringGT(), new GetName()));
      countries.forEach(System.out::println);System.out.println();
   }
}


class Country
{
   private String name;
   private double area;
   private double population;

   public Country(String name, double area, double population) {
      this.name = name;
      this.area = area;
      this.population = population;
   }

   public String getName() { return this.name; }

   public double getArea() { return this.area; }

   public double getPopulation() { return this.population; }

   public String toString() {
      return name + ": area = " + area + ", population = " + population;
   }
}
